import logging
import re
from typing import Any, AsyncIterator, List, Optional, Text, Tuple

import faker
from faker import Faker
from rasa.shared.importers.rasa import RasaFileImporter
from rasa.shared.nlu.training_data import loading
from rasa.shared.nlu.training_data.entities_parser import find_entities_in_training_example
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData

logger = logging.getLogger(__name__)


class PlaceholderImporter(RasaFileImporter):
    """
    PlaceholderImporter replaces placeholders with fake data. It enables you to write training data without worrying
    about other data like numbers, names, texts, etc.

    Usage in config.yml:

    ```yml
    importers:
        - name: rasam.PlaceholderImporter
          fake_data_count: 50
    ```

    If `fake_data_count` is not specified, the default value is 1.
    """

    DEFAULT_FAKE_DATA_COUNT = 1
    FAKE_MAP = {
        "integer": lambda faker_: faker_.pyint(min_value=-9999, max_value=9999),
        "decimal": lambda faker_: faker_.pyfloat(min_value=-9999, max_value=9999),
        "number": lambda faker_: getattr(faker_, faker_.random_choices(["pyint", "pyfloat"], length=1)[0])(
            min_value=-9999, max_value=9999
        ),
        "name": lambda faker_: faker_.name(),
        "first_name": lambda faker_: faker_.first_name(),
        "last_name": lambda faker_: faker_.last_name(),
        "text": lambda faker_: faker_.text(),
        "word": lambda faker_: faker_.word(),
        "paragraph": lambda faker_: faker_.paragraph(),
        "uri": lambda faker_: faker_.uri(),
        "url": lambda faker_: faker_.url(),
        "local_uri": lambda faker_: faker_.file_path(depth=faker_.random_int(min=1, max=5), extension="html"),
        "email": lambda faker_: faker_.email(),
        "date": lambda faker_: faker_.date(),
        "time": lambda faker_: faker_.time(),
        "month": lambda faker_: faker_.month_name(),
        "day": lambda faker_: faker_.day_of_week(),
        "timezone": lambda faker_: faker_.timezone(),
        "company": lambda faker_: faker_.company(),
        "license_plate": lambda faker_: faker_.license_plate(),
        "address": lambda faker_: faker_.address(),
        "city": lambda faker_: faker_.city(),
        "country": lambda faker_: faker_.country(),
        "user_agent": lambda faker_: faker_.user_agent(),
        "password": lambda faker_: faker_.password(),
        "user_name": lambda faker_: faker_.user_name(),
        "file_path": lambda faker_: faker_.file_path(),
    }

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        keys = "|".join(self.FAKE_MAP.keys())
        regex = f"(\\{{({keys}|any)\\}}|@({keys}|any)(?=([^\\w]|$|@)))"
        self.placeholder_regex = re.compile(regex)

    async def get_nlu_data(self, language: Optional[Text] = "en") -> TrainingData:
        fake_data_count = self.DEFAULT_FAKE_DATA_COUNT

        for importer in self.config["importers"]:
            if importer.get("name") == "rasam.PlaceholderImporter":
                fake_data_count = importer.get("fake_data_count", self.DEFAULT_FAKE_DATA_COUNT)

        faker_ = faker.Faker()
        faker_.seed_instance(fake_data_count)

        training_data = [loading.load_data(nlu_file, language) for nlu_file in self._nlu_files]

        new_training_data = []

        for data in training_data:
            training_examples = []
            example: Message
            for example in data.training_examples:
                if example.get("intent"):
                    matches = [i async for i in self.find_placeholders(example.data.get("text"))]
                    if matches:
                        async for new_message in self.replace_placeholders(example, faker_, matches, fake_data_count):
                            training_examples.append(new_message)
                    else:
                        training_examples.append(example)
                else:
                    training_examples.append(example)
            new_training_data.append(
                TrainingData(
                    training_examples, data.entity_synonyms, data.regex_features, data.lookup_tables, data.responses
                )
            )

        merged_training_data = TrainingData().merge(*new_training_data)
        merged_training_data._fill_response_phrases()
        return merged_training_data

    async def replace_placeholders(
        self, example: Message, faker_: Faker, matches: List[Tuple[Any, ...]], count: int
    ) -> AsyncIterator[Message]:
        original_text = await self.rebuild_original_text(example)
        for _ in range(count):
            text = await self.replace_placeholders_in_text(example.data.get("text"), faker_, matches)
            original_text = await self.replace_placeholders_in_text(original_text, faker_, matches)
            entities = find_entities_in_training_example(original_text)
            new_message = Message.build(text, example.get("intent"), entities)
            yield new_message

    async def replace_placeholders_in_text(self, text: str, faker_: Faker, matches: List[Tuple[Any, ...]]) -> str:
        """
        Replaces each placeholder in text from items in matches using the fake data.
        """
        for placeholder, faker_string, start in sorted(matches, key=lambda x: x[2], reverse=True):
            if faker_string == "any":
                faker_string = faker_.random_choices(list(self.FAKE_MAP.keys()), length=1)[0]
            text = text[:start] + text[start:].replace(placeholder, str(self.FAKE_MAP[faker_string](faker_)), 1)
        return text

    async def find_placeholders(self, text: str) -> AsyncIterator[Tuple[Any, ...]]:
        """
        Find placeholder names in text. See FAKE_MAP for valid placeholder names.

        Examples:
            hello {any}
            {name} went to {city}
            see you on @day at @time

        yields: Tuple(placeholder text, placeholder name, match start)
        """
        for item in self.placeholder_regex.finditer(text):
            yield tuple([i for i in item.groups() if i][:2]) + (item.start(),)

    @staticmethod
    async def rebuild_original_text(example: Message) -> str:
        """
        Rebuilds original training text in Markdown form.
        """
        original_entities = example.get("entities")
        original_text = example.data.get("text")
        if original_entities:
            original_text = list(original_text)
            for entity in sorted(original_entities, key=lambda x: x.get("start"), reverse=True):
                start = entity["start"]
                end = entity["end"]
                value = entity["value"]
                name = entity["entity"]
                original_text[start:end] = f"[{value}]({name})"
            original_text = "".join(original_text)
        return original_text



from rasam.components.regex import RegexEntityExtractor
from rasam.components.url import URLEntityExtractor
from rasam.decorators import action
from rasam.importers.placeholder import PlaceholderImporter



from typing import Callable, Text, Type

from rasa_sdk import Action


def action(f: Callable) -> Type[Action]:
    class RasamAction(Action):
        def name(self) -> Text:
            return f.__name__

    setattr(RasamAction, "run", f)
    return RasamAction



from typing import Any, Dict, Optional, Set, Text

import urlextract
from rasa.nlu.extractors.extractor import EntityExtractor
from rasa.shared.nlu.training_data.message import Message


class URLEntityExtractor(EntityExtractor):
    def __init__(self, component_config: Optional[Dict[Text, Any]] = None) -> None:
        super().__init__(component_config)
        self.extractor = urlextract.URLExtract()

    def process(self, message: Message, **kwargs: Any) -> None:
        urls: Set[Any] = set()
        last_pos = 0
        for url in self.extractor.gen_urls(message.data.get("text")):
            start = message.data.get("text").find(url, last_pos)
            end = start + len(url)
            last_pos = end
            urls.add(
                tuple(
                    {
                        "start": start,
                        "end": end,
                        "value": url,
                        "entity": "URL",
                        "extractor": self.name,
                        "confidence": 1.0,
                    }.items()
                )
            )
        entities = message.get("entities", []) + list(
            sorted(map(dict, urls), key=lambda x: x.get("start"))  # type: ignore
        )

        message.set(
            "entities",
            sorted(entities, key=lambda x: x.get("confidence", 0), reverse=True),
            add_to_output=True,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rasam', 'rasam.components', 'rasam.importers']

package_data = \
{'': ['*']}

install_requires = \
['Faker>=8.1.4,<9.0.0', 'rasa>=2.6.1,<3.0.0', 'urlextract>=1.2.0,<2.0.0']

setup_kwargs = {
    'name': 'rasam',
    'version': '0.4.0',
    'description': 'Rasa Improved',
    'long_description': '# rasam\n\nRasa Improved\n\n<table>\n    <tr>\n        <td>License</td>\n        <td><img src=\'https://img.shields.io/pypi/l/rasam.svg\' alt="License"></td>\n        <td>Version</td>\n        <td><img src=\'https://img.shields.io/pypi/v/rasam.svg\' alt="Version"></td>\n    </tr>\n    <tr>\n        <td>Travis CI</td>\n        <td><img src=\'https://www.travis-ci.com/roniemartinez/rasam.svg?branch=master\' alt="Travis CI"></td>\n        <td>Coverage</td>\n        <td><img src=\'https://codecov.io/gh/roniemartinez/rasam/branch/master/graph/badge.svg\' alt="CodeCov"></td>\n    </tr>\n    <tr>\n        <td>Supported versions</td>\n        <td><img src=\'https://img.shields.io/pypi/pyversions/rasam.svg\' alt="Python Versions"></td>\n        <td>Wheel</td>\n        <td><img src=\'https://img.shields.io/pypi/wheel/rasam.svg\' alt="Wheel"></td>\n    </tr>\n    <tr>\n        <td>Status</td>\n        <td><img src=\'https://img.shields.io/pypi/status/rasam.svg\' alt="Status"></td>\n        <td>Downloads</td>\n        <td><img src=\'https://img.shields.io/pypi/dm/rasam.svg\' alt="Downloads"></td>\n    </tr>\n</table>\n\n## Support\nIf you like `rasam` or if it is useful to you, show your support by buying me a coffee.\n\n<a href="https://www.buymeacoffee.com/roniemartinez" target="_blank"><img src="https://cdn.buymeacoffee.com/buttons/default-orange.png" alt="Buy Me A Coffee" height="41" width="174"></a>\n\n## Usage\n\n### Installation\n\n```bash\npip install rasam\n```\n\n### Rasa `config.yml`\n\n```yaml\nimporters:\n  - name: rasam.PlaceholderImporter\n    fake_data_count: 10  # default value is 1\n\npipeline:\n  - name: rasam.RegexEntityExtractor\n  - name: rasam.URLEntityExtractor\n```\n\n### Rasa `nlu.yml`\n\n#### PlaceholderImporter\n\nThe `PlaceholderImporter` removes the need to write unnecessary information (eg. name, address, numbers, etc.) and helps focus on writing test data.\n\n#### Using `{}` placeholder\n\n```yaml\nnlu:\n- intent: tell_name\n  examples: |\n    - My name is {name}\n    - I am {name} and he is {name}\n```\n\n#### Using `@` placeholder\n\n```yaml\nnlu:\n- intent: tell_address\n  examples: |\n    - I live in @address\n    - I stay at @address and @address\n```\n\n#### Mixing `{}` and `@` placeholders\n\nIt is possible to mix both `{}` and `@` placeholders but it is recommended to use only one style for consistency.\n\n#### Available placeholders\n\n- any (if you need just any data)    \n- integer    \n- decimal    \n- number     \n- name       \n- first_name \n- last_name  \n- text       \n- word       \n- paragraph  \n- uri        \n- url        \n- local_uri  \n- email      \n- date         \n- time         \n- month        \n- day          \n- timezone     \n- company      \n- license_plate\n- address\n- city\n- country\n- user_agent\n- password\n- user_name\n- file_path\n\n### Rasam decorators\n\nRasa relies too heavily on classes to define objects like actions, forms, etc. \nRasam aims to remove these Rasa boilerplates to make writing chatbots easier.\n\n#### @action decorator\n\nThe `@action` decorator converts function into an Action class. \nHere is an example of how we can write custom classes in Rasa:\n\n```python\nclass ActionHelloWorld(Action):\n\n    def name(self) -> Text:\n        return "action_hello_world"\n\n    def run(self, dispatcher: CollectingDispatcher,\n            tracker: Tracker,\n            domain: Dict[Text, Any]) -> List[Dict[Text, Any]]:\n\n        dispatcher.utter_message(text="Hello World!")\n\n        return []\n\n```\n\nThe above code can be simplified using Rasam\'s `@action` decorator.\n\n```python\nfrom rasam import action\n\n\n@action\ndef action_hello_world(\n    self: Action, dispatcher: CollectingDispatcher, tracker: Tracker, domain: Dict[Text, Any]\n) -> List[Dict[Text, Any]]:\n    dispatcher.utter_message(text="Hello World!")\n    return []\n```\n\n\n\n## Author\n[Ronie Martinez](ronmarti18@gmail.com) \n',
    'author': 'Ronie Martinez',
    'author_email': 'ronmarti18@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/roniemartinez/rasam',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.2,<3.9',
}


setup(**setup_kwargs)



import os
import re
from typing import Any, Dict, Optional, Set, Text

from rasa.nlu import utils
from rasa.nlu.components import Component
from rasa.nlu.config import RasaNLUModelConfig
from rasa.nlu.extractors.extractor import EntityExtractor
from rasa.nlu.model import Metadata
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.training_data.training_data import TrainingData
from rasa.shared.utils import io


class RegexEntityExtractor(EntityExtractor):
    def __init__(self, component_config: Optional[Dict[Text, Any]] = None) -> None:
        super().__init__(component_config)
        self.regex_features = []
        if not component_config:
            return
        for regex_feature in component_config.get("regex_features", []):
            regex_feature["compiled"] = re.compile(regex_feature["pattern"])
            self.regex_features.append(regex_feature)

    def train(
        self,
        training_data: TrainingData,
        config: Optional[RasaNLUModelConfig] = None,
        **kwargs: Any,
    ) -> None:
        self.regex_features = training_data.regex_features

    def process(self, message: Message, **kwargs: Any) -> None:
        matches: Set[Any] = set()
        for regex_feature in self.regex_features:
            for match in regex_feature["compiled"].finditer(message.data.get("text")):
                matches.add(
                    tuple(
                        {
                            "start": match.start(),
                            "end": match.end(),
                            "value": match.group(0),
                            "entity": regex_feature["name"],
                            "extractor": self.name,
                            "confidence": 1.0,
                        }.items()
                    )
                )
        entities = message.get("entities", []) + list(map(dict, matches))

        message.set(
            "entities",
            sorted(entities, key=lambda x: x.get("confidence", 0), reverse=True),
            add_to_output=True,
        )

    def persist(self, file_name: Text, model_dir: Text) -> Optional[Dict[Text, Any]]:
        if self.regex_features:
            file_name = file_name + ".json"
            utils.write_json_to_file(os.path.join(model_dir, file_name), self.regex_features)
            return {"file": file_name}
        return {"file": None}

    @classmethod
    def load(
        cls,
        meta: Dict[Text, Any],
        model_dir: Text,
        model_metadata: Optional["Metadata"] = None,
        cached_component: Optional["Component"] = None,
        **kwargs: Any,
    ) -> "Component":
        file_name = meta.get("file")
        regex_features = []
        if file_name:
            regex_features = io.read_json_file(os.path.join(model_dir, file_name))
        meta["regex_features"] = regex_features
        return cls(meta)
